/******************************************************************************
 *
 * Licensed Materials - Property of IBM
 *
 * (C) COPYRIGHT IBM Corp. 2005
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 *
 *****************************************************************************/

/*********************************************************************
 * FILE: %Z%%M%  %I%  %W% %G% %U%
 *
 * Description:
 *
 * This servlet authenticates a user by doing a JAAS login to TIM.
 * It also detects if a user needs to update his challenge-response
 * answers.
 *
 *********************************************************************/

package examples.expi;

import java.io.*;
import java.rmi.RemoteException;
import java.util.*;

import javax.security.auth.Subject;
import javax.security.auth.login.*;
import javax.servlet.ServletException;
import javax.servlet.http.*;

import com.ibm.itim.apps.*;
import com.ibm.itim.apps.identity.*;
import com.ibm.itim.apps.jaas.callback.PlatformCallbackHandler;
import com.ibm.itim.apps.jaas.spi.PlatformPrincipal;

/**
 * @version 	1.0
 * @author
 */
public class logonServlet extends HttpServlet {
	private static String LOGON_CONTEXT_ID;
	private static String LOGON_PAGE;
	private static String HOME_PAGE;
	private static String SELF_CHANGEPWD_PAGE;

	private boolean bSSOEnabled = false;
	private expiUtil utilObject = null;

	/**
	 * Method init
	 * Initializes several class variables from the properties file.
	 * @see javax.servlet.GenericServlet#init()
	 */
	public void init() throws ServletException {

		log("EXPI:logonServlet.init(): begin");

		utilObject = new expiUtil();
		if (utilObject != null && utilObject.isProperties()) {
			this.LOGON_CONTEXT_ID = utilObject.getLoginContextID();
			log("EXPI:init(): LOGON_CONTEXT_ID = " + LOGON_CONTEXT_ID);

			this.LOGON_PAGE = utilObject.getPropertySSOCheck(utilObject.LOGON_PAGE);
			log("EXPI:init(): LOGON_PAGE = " + LOGON_PAGE);

			this.HOME_PAGE = utilObject.getProperty(utilObject.HOME_PAGE);
			log("EXPI:init(): HOME_PAGE = " + HOME_PAGE);

			bSSOEnabled = utilObject.isSSOEnabled();
			log("EXPI:init(): bSSOEnabled = " + bSSOEnabled);

			this.SELF_CHANGEPWD_PAGE = utilObject.getProperty(utilObject.SELF_CHANGEPWD_PAGE);
			log("EXPI:init(): SELF_CHANGEPWD_PAGE = " + SELF_CHANGEPWD_PAGE);
		}

		log("EXPI:init(): end");

	} // init

	/**
	* @see javax.servlet.http.HttpServlet#void (javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
	*/
	public void doGet(HttpServletRequest req, HttpServletResponse resp)
		throws ServletException, IOException {
		req.setCharacterEncoding("UTF-8");

		log("EXPI: logonServlet.doGet()");

		// Initially always set the fact that iv_user is not the only indicator of SSO being enabled
		// The property file(s) must be configured accordingly to force SSO.

		if (bSSOEnabled)
			req.setAttribute(expiUtil.SSO_ENABLED,"yes");
		else
			req.setAttribute(expiUtil.SSO_ENABLED,"no");

		// check if we found a properties file
		if (utilObject == null || !utilObject.isProperties()) {
			log("EXPI:doPost(): Could not find properties file");
			utilObject.forward(
				req,
				resp,
				"Could not initialize: Property file missing?",
				LOGON_PAGE);
			return;
		}

		// if page is the result of logging out invalidate session

		HttpSession session = req.getSession(false);
		if (session != null) {
			/** Check to see if sso is enabled.
			 * If so...check if the iv-user is provided
			 * if the iv_user is provided check to see if they are already
			 * authenticated to ITIM (if so...forward to the home page), otherwise
			 * invalidate the session and carry out the authentication.
			*/
			if (bSSOEnabled) {
				String sessionID =
					(String) session.getAttribute(
						expiUtil.SSO_SESSION_LOGON_ID);
				String userID = (String) req.getHeader("iv-user");

				if (sessionID != null && userID != null) {
					log(
						"EXPI: [logonServlet.doGet()] sessionID = "
							+ sessionID
							+ " userID = "
							+ userID);
					if (sessionID.equals(userID)) {
						/** SSO user already supplied an authenticated - forward to the home page
						 */
						log("EXPI: [logonServlet.doGet()] SSO user already authenticated");
						utilObject.forward(
							req,
							resp,
							"",
							utilObject.getProperty(expiUtil.HOME_PAGE));
					} else {
						/** different user from previous session so invalidate
						 * and reauthenticate them
						 */

						session.invalidate();
						session = req.getSession(true);
						log("EXPI: [logonServlet.doGet()] SSO invalidating previous session ");

						doPost(req, resp);

						utilObject.forward(
							req,
							resp,
							"",
							utilObject.getProperty(expiUtil.HOME_PAGE));
					}
					return;
				} else if (userID == null) {
					/** the SSO user is not yet authenticated so hand the request off to the
					 * doPost routine which will handle the normal authentication process
					 */

					// Invalidate the session; Need new session to do the forward!
					session.invalidate();
					session = req.getSession(true);

					log("EXPI: [logonServlet.doGet()] SSO - no user ID specified - logon aborted");
					utilObject.forward(
						req,
						resp,
						"SSO Logon Failed - no user ID specified - logon aborted",
						utilObject.getProperty(expiUtil.SSO_ERROR_PAGE));
					return;
				}
			} else {
				session.invalidate();
				session = req.getSession(true);  // For error handling
			}

			log("EXPI: [logonServlet.doGet()] invalidating previous session");

		} else {
			if (bSSOEnabled) {
				String userID = (String) req.getHeader("iv-user");

				// create a session for error handling control.

				session = req.getSession(true);
				if (userID == null || userID.equals("")) {
					utilObject.forward(
						req,
						resp,
						"SSO Logon enabled - no WebSEAL user ID supplied",
						utilObject.getProperty(expiUtil.SSO_ERROR_PAGE));
					return;
				}

				log("EXPI: [logonServlet.doGet()] calling doPost for SSO logon");
				doPost(req, resp);
				return;
			}
		}

		log("EXPI: [logonServlet.doGet()] default - calling LOGON_PAGE");

		utilObject.forward(
			req,
			resp,
			"",
			utilObject.getProperty(expiUtil.LOGON_PAGE));
	}

	/**
	* @see javax.servlet.http.HttpServlet#void (javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
	*/
	public void doPost(HttpServletRequest req, HttpServletResponse resp)
		throws ServletException, IOException {
		req.setCharacterEncoding("UTF-8");

		String password;
		String userID;
		String tenantID;
		String errorMsg = "";
		ITIMFailedLoginException failedLoginException = null;
		boolean bError = false;
		LoginContext lc;
		PlatformCallbackHandler handler = null;

		log("EXPI:logonServlet.doPost()");

		/**
		 * Need a session whether the login was successful or not!
		 * If unsuccessful login, need a session to hold the error info.
		 */
		HttpSession session = req.getSession(false);

		if (session == null) {
			session = req.getSession(true);
		}

		session.setAttribute(expiUtil.LOGGED_ON, "");

		// check if we found a properties file
		if (utilObject == null || utilObject.isProperties() == false) {
			log("EXPI:doPost(): Could not find properties file");
			utilObject.forward(
				req,
				resp,
				"Could not initialize: Property file missing?",
				LOGON_PAGE);
			return;
		}

		try {
			bError = false;

			/*
			 * display the parameter list if desired
			 * NOTE: care should be taken with passwords and such -
			 * code not included here.
			 *
			Enumeration enum = req.getParameterNames();
			log("EXPI: Parameter list");

			while (enum.hasMoreElements()) {
				String element = (String) enum.nextElement();
				String eValue = req.getParameter(element);
				log("EXPI: " + element + ": " + eValue);
			}
			*/

			// set the user ID in session attributes to allow other servlets
			// access to this information.  This includes the expired password jsp!

			if (bSSOEnabled) {
				userID = (String) req.getHeader("iv-user");

				/** set this to allow other pages to know we are using SSO
				 * For example: The header generation conditionally
				 * outputs the logout link.
				 */

				session.setAttribute(expiUtil.SSO_ENABLED, "yes");
			}
			else
				userID = (String) req.getParameter(expiUtil.LOGON_ID);

			if ((userID == null) || (userID.trim().length() == 0)){
				throw new ITIMFailedLoginException(
					ITIMFailedLoginException.NO_USER_ID);
			}
			session.setAttribute(expiUtil.LOGON_ID, userID);

			/** if SSO is enabled set the SSO Session Logon ID
			 * the to the current userID
			 */

			if (bSSOEnabled) {
				session.setAttribute(expiUtil.SSO_SESSION_LOGON_ID, userID);
			}

			// Setup work
			PlatformContext platform = utilObject.createPlatformContext();

			handler =
				createPlatformCallbackHandler(platform, req);
			log("EXPI:doPost(): calling LoginContext.");

			// Try to login to TIM
			lc = new LoginContext(utilObject.getLoginContextID(), handler);
			lc.login();
			log("EXPI:doPost(): login was successful");

			// User is authenticated -- setup the session
			session.setAttribute(expiUtil.LOGGED_ON, "yes");
			session.setAttribute(expiUtil.PLATFORM_CONTEXT, platform);

			Subject subject = lc.getSubject();
			session.setAttribute(expiUtil.SUBJECT, subject);

			String userDN = (PlatformPrincipal.getPrincipal(subject)).getName();
			log("EXPI:doPost(): Principal userDN " + userDN);

			// Get the PersonMO object that corresponds to the logged in user

			PersonMO personMo = utilObject.lookupPerson(platform, subject);
			session.setAttribute(expiUtil.PERSONMO, personMo);

			// See if the user needs to update their challenge response answers
			// This sets msgs in the session as needed

			checkChallengeResponse(session, platform, subject);
			resp.sendRedirect("mainServlet");

		// NOTE: CredentialExpiredException must be caught before LoginException
		} catch (CredentialExpiredException e) {
			session.setAttribute("passwordExpired", new Boolean(true));
			log("doPost(): Caught CredentialExpiredException, forwarding to expiredpassword page");
			utilObject.forward(req, resp, errorMsg, SELF_CHANGEPWD_PAGE);
			return;
		} catch (LoginException e) {
			e.printStackTrace ();
			String eMsg ="";
			if ( e instanceof ITIMFailedLoginException) {
				eMsg = ((ITIMFailedLoginException)e).getMessageId();
			} else {
				eMsg = e.getMessage();
			}
			if (eMsg == null || eMsg.equals("")){
				errorMsg = "An unknown error occurred";
			} else {
				if (eMsg.equals(ITIMFailedLoginException.INVALID_PASSWORD)) {
					//code to retrieve and check the number of attempts
					//remaining
					int remainingAttempts = -1;
					try {
						remainingAttempts = handler.getRemainingAttempts();
					} catch(ApplicationException ex) {
						errorMsg = ex.getMessage ();
					}
					if (remainingAttempts > -1) {
						failedLoginException =
						new ITIMFailedLoginException(
						ITIMFailedLoginException.INVALID_PASSWORD_ATTEMPTS,
						new Object[]{Integer.toString(remainingAttempts)});

						errorMsg = failedLoginException.getMessage();
					}
				}

				if (errorMsg.equals("")){
					 errorMsg = e.getMessage ();
				}
			}
			log("doPost: Caught LoginException: " + errorMsg);
			bError = true;
		} catch (ApplicationException e) {
         log("doPost: Caught ApplicationException: " + e.getMessage());
			e.printStackTrace();
			bError = true;
			errorMsg = "Sign on failed: could not establish a connection to the server.";
		} catch (RemoteException e) {
			e.printStackTrace();
         log("doPost: Caught RemoteException: " + e.getMessage());
			bError = true;
			errorMsg = "Sign on failed: could not establish a connection to the server.";
		}

		if (bError) {
			log("doPost(): Invalidating current session; starting fresh");
			session.invalidate();
			session = req.getSession(true);
			utilObject.forward(req, resp, errorMsg, LOGON_PAGE);
		}

	} // doPost

	/**
	 * Method checkChallengeResponse.
	 *
	 * Determine whether the user who is logging in needs to
	 * update his challenge response answers.  This should be true:
	 * - The first time a user logs on
	 * - When an administrator makes a change to the system-wide
	 *   challenge response settings that causes the user to update his
	 *   settings
	 *
	 * Update the session with any needed messages.
	 * These will be used by main.jsp.
	 * @param session
	 * @param platform
	 * @param subject
	 */
	private void checkChallengeResponse(
		HttpSession session,
		PlatformContext platform,
		Subject subject) {

		try {
			ChallengeResponseManager crm =
				new ChallengeResponseManager(platform, subject);

			boolean changeNeeded = crm.isEnforceChallengeResponse();
			log("EXPI: checkChallengeResponse(): challenge response updated needed = "
				+ changeNeeded);

			// Need to set to keep the session up-to-date
			session.setAttribute("updateCRAnswers", new Boolean(changeNeeded));
		} catch (Exception e) {
			log(
				"EXPI: checkChallengeResponse(): Caught an exception: "
					+ e.getMessage());
			session.setAttribute("errorMessage", e.getMessage());
		}

	} // checkChallengeResponse

	/**
	 * Method createPlatformCallbackHandler.
	 * Returns the PlatformCallbackHandler for the PlatformContext
	 * (as specified by the properties file) used to logon the user
	 * via the JAAS interface.
	 *
	 * A different constructor is used depending on whether sign-on (SSO) is
	 * enabled (via a property file setting).
	 *
	 * @param platform
	 * @param req
	 * @return PlatformCallbackHandler - return null if no user ID was specified or
	 * no password was supplied for a non SSO logon attempt.
	 */
	public PlatformCallbackHandler createPlatformCallbackHandler(
		PlatformContext platform,
		HttpServletRequest req) {

		log("EXPI:createPlatformCallbackHandler(): begin");

		PlatformCallbackHandler handler = null;

		// if the properties file was setup with ssoEnabled then we expect to find
		// the user in the specified logonID

		if (bSSOEnabled) {
			String userID = (String) req.getHeader("iv-user");
			log("EXPI:createPlatformCallbackHandler(): SSO logon userID = " + userID);

			handler =
				new PlatformCallbackHandler(
					utilObject.getTenantID(),
					userID,
					Locale.US);
			// Samples assume English
		} else {
			String password = (String) req.getParameter("password");
			String userID = (String) req.getParameter(expiUtil.LOGON_ID);
			log("EXPI:createPlatformCallbackHandler(): " + expiUtil.LOGON_ID + " = " + userID);
			handler =
				new PlatformCallbackHandler(
					utilObject.getTenantID(),
					userID,
					password);
		}

		handler.setPlatformContext(platform);

		log("EXPI:createPlatformCallbackHandler(): end");

		return handler;
	} // createPlatformCallbackHandler

}
